import rclpy
from rclpy.node import Node
from rclpy.node import Node,SetParametersResult,Parameter
from std_msgs.msg import  String
import soundcard as sc
from io import BytesIO
import subprocess,os
import pyttsx4
from dataclasses import dataclass
from typing import List
import gc
import threading

from lifecycle_msgs.msg import State
from lifecycle_msgs.msg import Transition
from tts.lifecycle import LifecycleNode

from tts.utils import  (
    VOLUME_CONSTANTS,   
    JETSON_SPEAKER,
    raw_wav2byte,
    wavfile2byte,
    byte2np,
    combine_audio_samples,
    split_en_and_zh,
)

@dataclass  
class Config:
    topic: str
    volume: float
    platform: str


class TextToSpeech(LifecycleNode):
    def __init__(self):
        super().__init__('tts_node')
        self.declare_parameters(
            namespace='',
            parameters=[(attr, None) for attr in Config.__annotations__]
        )
        self.logger = self.get_logger()  
        self.config: Config = None
        self.lock = threading.Lock()
        self.add_on_set_parameters_callback(self.config_update)
        
    def on_configure(self):
        self.logger.info("on_configure() is called")
        self.config = Config(
            **{attr: self.get_parameter(attr).value for attr in Config.__annotations__}
        ) if self.config is None else self.config
        
        self.speaking = False
        try:
            with self.lock:
                self.configure_speaker()
        except Exception as e:
            self.logger.error(f"Init failed: {e}")
        return Transition.TRANSITION_CALLBACK_SUCCESS
    
    def on_cleanup(self):
        self.logger.info("on_cleanup() is called")
        return Transition.TRANSITION_CALLBACK_SUCCESS
    
    def on_activate(self):        
        self.logger.info("on_activate() is called")
        self.sub = self.create_subscription(String,self.config.topic, self.speak, 10)
        return Transition.TRANSITION_CALLBACK_SUCCESS
    
    def on_deactivate(self):
        self.logger.info("on_deactivate() is called")
        # self.sub.destroy()
        self.sub = None
        
        gc.collect()
        return Transition.TRANSITION_CALLBACK_SUCCESS

    # def destroy_node(self) -> bool:
    #     return super().destroy_node()
    
    def config_update(self, params: List[Parameter]):
        self.logger.info(f'Update TTS: {[str(p.name) for p in params]}')
        for param in params:
            if param.name in self.config.__annotations__:
                if param.value == self.config.__getattribute__(param.name):
                    continue
                self.config.__setattr__(param.name, param.value)
        return SetParametersResult(successful=True)


    def configure_speaker(self):
        self.volume_constant = VOLUME_CONSTANTS[self.config.platform]
        if self.config.platform == 'onnx' or self.config.platform == 'bpu':
            try:
                self.speaker = sc.default_speaker()
            except Exception as e:
                raise Exception(f"TTS init failed: {e}")
            
            # voices_choices = {
            #     12:'Mandarin',
            #     19:'English_Brit',
            #     24:'English_USA',
            #     108:'Cantonese',
            # }
            self.voice = 108
        else:
            try:
                self.speaker = sc.get_speaker(JETSON_SPEAKER)
            except Exception as e:
                raise Exception(f"TTS init failed: {e}")
            # voices_choices = {
            #     11:'Mandarin',
            #     18:'English_Brit',
            #     23:'English_USA',
            #     99:'Cantonese',
            # }
            self.voice = 99
        try:
            self.chinese_engine = pyttsx4.init()
            voices = self.chinese_engine.getProperty('voices')
            # set to cantonese, as it is good in both english and cantonese
            self.chinese_engine.setProperty('voice', voices[self.voice].id)
            # speed up the voice
            rate = self.chinese_engine.getProperty('rate')
            self.chinese_engine.setProperty('rate', int(rate*1.2))
        except Exception as e:
            raise Exception(f"TTS init failed: {e}")
        
    def speak(self, msg : String):
        text = msg.data
        if text == '' or self.speaking:
            return 
        self.logger.info(f"Speak: {text}")
        self.speaking = True
        try:
            en_text = text
            # text_split = self.split_en_and_zh(text)
            # en_text = text_split.get('en', '')
            en_audio_data, en_sample_rate = self.speak_eng(en_text)
            # zh_text = text_split.get('zh', '')
            # zh_audio_data, zh_sample_rate = self.speak_zh(zh_text)   
            audio_data, sample_rate = combine_audio_samples(
                [en_audio_data],
                [en_sample_rate]
            )
            if len(audio_data) > 0:
                normalized_samples = byte2np(audio_data, self.config.volume, self.volume_constant)
                self.speaker.play(normalized_samples, samplerate=sample_rate)
        except Exception as e:
            self.logger.warning(f'TTS failed, {e}')
        self.speaking = False        
        
    def speak_eng(self, text):
        if text != '':
            file = '/tmp/test.wav'
            subprocess.call(['pico2wave', f'-w={file}', text])
            en_audio_data, en_sample_rate = wavfile2byte(file)
            os.remove(file)
            return en_audio_data, en_sample_rate
        return None, None
    
    def speak_zh(self, text):
        if text != '':
            wav_file = BytesIO()
            self.chinese_engine.save_to_file(text, wav_file)
            self.chinese_engine.runAndWait()
            zh_audio_data, zh_sample_rate = raw_wav2byte(wav_file)
            return zh_audio_data, zh_sample_rate
        return None, None
    

def main(args=None):
    rclpy.init(args=args)
    node = TextToSpeech()
    try:
        rclpy.spin(node)
    except KeyboardInterrupt:
        pass
    rclpy.shutdown()